import sys
import os
import json
import zipfile
import base64
import pathlib

# Name a folder of unencrypted dindex and dlist to check for missing or duplicated blocks.
# This does not yet check dblock files for their presence, or their matching their dindex.
file_folder = sys.argv[1]

dlist_blocklist_set = set()
dlist_blockset_set = set()
dlist_block_set = set()
dlist_meta_block_set = set()
file_root = pathlib.Path(file_folder)
for entry in file_root.glob("*.dlist.zip"):
    with zipfile.ZipFile(entry) as dlist:
        filelist = json.load(dlist.open('filelist.json'))
        for file in filelist:
            dlist_meta_block_set.add(file["metahash"])   # Doesn't support metablocklists.
            if file["type"] != "File":
                continue
            if "blocklists" in file:
                dlist_blockset_set.add(file["hash"])
                for blocklist in file["blocklists"]:
                    dlist_blocklist_set.add(blocklist)
            else:
                dlist_block_set.add(file["hash"])

#print(dlist_blocklist_set)
#print(dlist_block_set)

file_root = pathlib.Path(file_folder)

dindex_blocklist_set = set()
dindex_blocklist_block_set = set()
for entry in file_root.glob("*.dindex.zip"):
#    print(entry, "blocklist cache being read")
    root = zipfile.Path(entry, "list/")
    for path in root.iterdir():
#        print(path.name, "blocklist hash seen")
        # Convert to regular base64 from URL/file safe alphabet which replaces +/ with -_
        blocklist = base64.b64encode(base64.urlsafe_b64decode(path.name)).decode()
        if blocklist not in dlist_blocklist_set:
            continue    # A stale blocklist could lack some of its blocks, so ignore this.
        if blocklist in dindex_blocklist_set and blocklist in dlist_blocklist_set:
            pass
            print("Duplicate blocklist", blocklist)   # Why these even from just a backup?
        else:
            dindex_blocklist_set.add(blocklist)
            blocklist_file = path.open("rb")
            while blockhash := blocklist_file.read(32):
                blockname = str(base64.b64encode(blockhash), "utf-8")
#                print(blockname, "added from blocklist")        
                dindex_blocklist_block_set.add(blockname)

dindex_block_set = set()          
for entry in file_root.glob("*.dindex.zip"):
#     print(entry, "volume index being read")
    root = zipfile.Path(entry, "vol/")
    for path in root.iterdir():
#            print(path.name, "volume index being read")
        for block in json.load(path.open())["blocks"]:
            hash = block["hash"]
#                print(hash, "seen in blocks list")
            if hash in dindex_block_set and (hash in dlist_block_set or hash in dlist_blocklist_set or hash in dindex_blocklist_block_set):
                pass
                print("Duplicate block", hash)
            else:
                dindex_block_set.add(hash)

#print(dindex_blocklist_set)
#print(dindex_blocklist_block_set)
#print(dindex_block_set)

print(len(dindex_block_set), "blocks seen in dindex files")
print(len(dindex_blocklist_set), "blocklists seen in dindex files")
if missing := dindex_blocklist_set - dindex_block_set:
    print("Missing", missing)
print(len(dlist_blocklist_set), "blocks used by dlist files blocklists")
if missing := dlist_blocklist_set - dindex_blocklist_set:
    print("Missing", missing)
print(len(dindex_blocklist_block_set), "blocks used by dlist large files data")
if missing := dindex_blocklist_block_set - dindex_block_set:
    print("Missing", missing)
print(len(dlist_block_set), "blocks used by dlist small files data")
if missing := dlist_block_set - dindex_block_set - {"47DEQpj8HBSa+/TImW+5JCeuQeRkm5NMpJWZG3hSuFU="}:
    print("Missing", missing)
print(len(dlist_meta_block_set), "blocks used by dlist files metadata")
if missing := dlist_meta_block_set - dindex_block_set:
    print("Missing", missing)
print(len(dindex_block_set - dlist_block_set - dlist_blocklist_set - dindex_blocklist_block_set - dlist_meta_block_set), "blocks unused")
print(len(dlist_blockset_set), "large blocksets in dlist files")
print(len(dlist_block_set), "small blocksets in dlist files")
print("small file blocksets that are also metadata blocksets:", dlist_block_set.intersection(dlist_meta_block_set))
print("small file blocksets that are also blocklists:", dlist_block_set.intersection(dindex_blocklist_set))